/*
 * Library to access the Windows Hibernation File (hiberfil.sys) format
 *
 * Copyright (C) 2012-2021, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#if !defined( _LIBHIBR_H )
#define _LIBHIBR_H

#include <libhibr/codepage.h>
#include <libhibr/definitions.h>
#include <libhibr/error.h>
#include <libhibr/extern.h>
#include <libhibr/features.h>
#include <libhibr/types.h>

#include <stdio.h>

#if defined( LIBHIBR_HAVE_BFIO )
#include <libbfio.h>
#endif

#if defined( __cplusplus )
extern "C" {
#endif

/* -------------------------------------------------------------------------
 * Support functions
 * ------------------------------------------------------------------------- */

/* Returns the library version
 */
LIBHIBR_EXTERN \
const char *libhibr_get_version(
             void );

/* Returns the access flags for reading
 */
LIBHIBR_EXTERN \
int libhibr_get_access_flags_read(
     void );

/* Returns the access flags for reading and writing
 */
LIBHIBR_EXTERN \
int libhibr_get_access_flags_read_write(
     void );

/* Returns the access flags for writing
 */
LIBHIBR_EXTERN \
int libhibr_get_access_flags_write(
     void );

/* Retrieves the narrow system string codepage
 * A value of 0 represents no codepage, UTF-8 encoding is used instead
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_get_codepage(
     int *codepage,
     libhibr_error_t **error );

/* Sets the narrow system string codepage
 * A value of 0 represents no codepage, UTF-8 encoding is used instead
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_set_codepage(
     int codepage,
     libhibr_error_t **error );

/* Determines if a file contains a HIBR file signature
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_check_file_signature(
     const char *filename,
     libhibr_error_t **error );

#if defined( LIBHIBR_HAVE_WIDE_CHARACTER_TYPE )

/* Determines if a file contains a HIBR file signature
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_check_file_signature_wide(
     const wchar_t *filename,
     libhibr_error_t **error );

#endif /* defined( LIBHIBR_HAVE_WIDE_CHARACTER_TYPE ) */

#if defined( LIBHIBR_HAVE_BFIO )

/* Determines if a file contains a HIBR file signature using a Basic File IO (bfio) handle
 * Returns 1 if true, 0 if not or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_check_file_signature_file_io_handle(
     libbfio_handle_t *bfio_handle,
     libhibr_error_t **error );

#endif /* defined( LIBHIBR_HAVE_BFIO ) */

/* -------------------------------------------------------------------------
 * Notify functions
 * ------------------------------------------------------------------------- */

/* Sets the verbose notification
 */
LIBHIBR_EXTERN \
void libhibr_notify_set_verbose(
      int verbose );

/* Sets the notification stream
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_notify_set_stream(
     FILE *stream,
     libhibr_error_t **error );

/* Opens the notification stream using a filename
 * The stream is opened in append mode
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_notify_stream_open(
     const char *filename,
     libhibr_error_t **error );

/* Closes the notification stream if opened using a filename
 * Returns 0 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_notify_stream_close(
     libhibr_error_t **error );

/* -------------------------------------------------------------------------
 * Error functions
 * ------------------------------------------------------------------------- */

/* Frees an error
 */
LIBHIBR_EXTERN \
void libhibr_error_free(
      libhibr_error_t **error );

/* Prints a descriptive string of the error to the stream
 * Returns the number of printed characters if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_error_fprint(
     libhibr_error_t *error,
     FILE *stream );

/* Prints a descriptive string of the error to the string
 * The end-of-string character is not included in the return value
 * Returns the number of printed characters if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_error_sprint(
     libhibr_error_t *error,
     char *string,
     size_t size );

/* Prints a backtrace of the error to the stream
 * Returns the number of printed characters if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_error_backtrace_fprint(
     libhibr_error_t *error,
     FILE *stream );

/* Prints a backtrace of the error to the string
 * The end-of-string character is not included in the return value
 * Returns the number of printed characters if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_error_backtrace_sprint(
     libhibr_error_t *error,
     char *string,
     size_t size );

/* -------------------------------------------------------------------------
 * File functions
 * ------------------------------------------------------------------------- */

/* Creates a file
 * Make sure the value file is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_initialize(
     libhibr_file_t **file,
     libhibr_error_t **error );

/* Frees a file
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_free(
     libhibr_file_t **file,
     libhibr_error_t **error );

/* Signals a file to abort its current activity
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_signal_abort(
     libhibr_file_t *file,
     libhibr_error_t **error );

/* Opens a file
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_open(
     libhibr_file_t *file,
     const char *filename,
     int access_flags,
     libhibr_error_t **error );

#if defined( LIBHIBR_HAVE_WIDE_CHARACTER_TYPE )

/* Opens a file
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_open_wide(
     libhibr_file_t *file,
     const wchar_t *filename,
     int access_flags,
     libhibr_error_t **error );

#endif /* defined( LIBHIBR_HAVE_WIDE_CHARACTER_TYPE ) */

#if defined( LIBHIBR_HAVE_BFIO )

/* Opens a file using a Basic File IO (bfio) handle
 * Returns 1 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_open_file_io_handle(
     libhibr_file_t *file,
     libbfio_handle_t *file_io_handle,
     int access_flags,
     libhibr_error_t **error );

#endif /* defined( LIBHIBR_HAVE_BFIO ) */

/* Closes a file
 * Returns 0 if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_close(
     libhibr_file_t *file,
     libhibr_error_t **error );

/* Reads (media) data at the current offset
 * Returns the number of bytes read or -1 on error
 */
LIBHIBR_EXTERN \
ssize_t libhibr_file_read_buffer(
         libhibr_file_t *file,
         void *buffer,
         size_t buffer_size,
         libhibr_error_t **error );

/* Reads (media) data at a specific offset
 * Returns the number of bytes read or -1 on error
 */
LIBHIBR_EXTERN \
ssize_t libhibr_file_read_buffer_at_offset(
         libhibr_file_t *file,
         void *buffer,
         size_t buffer_size,
         off64_t offset,
         libhibr_error_t **error );

/* Seeks a certain offset of the (media) data
 * Returns the offset if seek is successful or -1 on error
 */
LIBHIBR_EXTERN \
off64_t libhibr_file_seek_offset(
         libhibr_file_t *file,
         off64_t offset,
         int whence,
         libhibr_error_t **error );

/* Retrieves the current offset of the (media) data
 * Returns the offset if successful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_get_offset(
     libhibr_file_t *file,
     off64_t *offset,
     libhibr_error_t **error );

/* Retrieves the media size
 * Returns the 1 if succesful or -1 on error
 */
LIBHIBR_EXTERN \
int libhibr_file_get_media_size(
     libhibr_file_t *file,
     size64_t *media_size,
     libhibr_error_t **error );

#if defined( __cplusplus )
}
#endif

#endif /* !defined( _LIBHIBR_H ) */

